"use strict";

var logger = require('./logger');

var Functions = require('./Functions');
var SecurityError = require('./SecurityError');
var ValidationError = require('./ValidationError');
var GuiMessageError = require('./gui/GuiMessageError');
var NetworkError = require('./NetworkError');

/**
 * This static global object contains set of functions that returns/set data in
 * the Gui (html).
 */
function GuiConnector() {
  // X coordinate of the mouse in a browser.
  this.xPos = 0;
  // coordinate of the mouse in a browser.
  this.yPos = 0;

  this.getParams = [];
}

function returnThisOrSingleton(object) {
  if (object === undefined || object === null) {
    return GuiConnector.singleton;
  } else {
    return object;
  }
}

/**
 * List of GET params passed via url.
 */

GuiConnector.prototype.init = function () {
  var self = returnThisOrSingleton(this);

  if (!String.prototype.endsWith) {
    String.prototype.endsWith = function (pattern) {
      var d = this.length - pattern.length;
      return d >= 0 && this.lastIndexOf(pattern) === d;
    };
  }
  var isIE = /* @cc_on!@ */false || !!document.documentMode;

  if (isIE) {
    alert("This web page works well with Chrome, Firefox and Safari.");
  }
  // bootstrap tab initialization
  $("ul.nav-tabs a").click(function (e) {
    e.preventDefault();
    $(this).tab('show');
  });

  self.getParams = [];

  // find GuiConnector.getParams
  window.location.search.replace(/\??(?:([^=]+)=([^&]*)&?)/g, function () {
    function decode(s) {
      return decodeURIComponent(s.split("+").join(" "));
    }

    self.getParams[decode(arguments[1])] = decode(arguments[2]);
  });

  // force browser to update mouse coordinates whenever mouse move
  jQuery(document).ready(function () {
    $(document).mousemove(function (e) {
      self.updateMouseCoordinates(e.pageX, e.pageY);
    });
  });

  if (self._windowResizeEvents === undefined) {
    self._windowResizeEvents = [];

    if (window.onresize !== null && window.onresize !== undefined) {
      self.addWindowResizeEvent(window.onresize);
    }

    window.onresize = function () {
      for (var i = 0; i < self._windowResizeEvents.length; i++) {
        self._windowResizeEvents[i]();
      }
    };
  }
};

GuiConnector.prototype.addWindowResizeEvent = function (handler) {
  this._windowResizeEvents.push(handler);
};

GuiConnector.prototype.removeWindowResizeEvent = function (handler) {
  var events = this._windowResizeEvents;
  var index = events.indexOf(handler);
  if (index > -1) {
    events.splice(index, 1);
  } else {
    logger.warn("Cannot find listener", handler);
  }
};

/**
 * Returns name of the file with LCSB logo.
 *
 * @param bigLogo
 *          {@link Boolean} value determining if we want to have big logo or
 *          small one
 */
GuiConnector.prototype.getLcsbLogoImg = function (bigLogo) {
  if (bigLogo) {
    return 'lcsb_logo_mid.png';
  } else {
    return 'lcsb_logo.png';
  }
};

/**
 * Returns name of the file with image that should be presented when we are
 * waiting for data to be loaded.
 */
GuiConnector.prototype.getLoadingImg = function () {
  return "icons/ajax-loader.gif";
};

/**
 * Returns home directory for images in the application.
 */
GuiConnector.prototype.getImgPrefix = function () {
  return "resources/images/";
};

/**
 * Updates coordinates of the mouse in the browser.
 */
GuiConnector.prototype.updateMouseCoordinates = function (x, y) {
  var self = returnThisOrSingleton(this);
  self.xPos = x;
  self.yPos = y;
};

GuiConnector.prototype.showProcessing = function (messageText) {
  var self = returnThisOrSingleton(this);
  if (self._processingDialog === undefined) {
    self._processingDialog = document.createElement("div");
    self._processingDialogContent = document.createElement("div");
    self._processingDialog.appendChild(self._processingDialogContent);
    document.body.appendChild(self._processingDialog);
    $(self._processingDialog).dialog({
      modal: true,
      title: "PROCESSING",
      width: "150px",
      closeOnEscape: false,
      dialogClass: 'minerva-no-close'
    });
  }
  if (messageText === undefined) {
    messageText = "PROCESSING";
  }
  var messageImg = Functions.createElement({
    type: "img",
    src: 'resources/images/icons/ajax-loader.gif'
  });
  self._processingDialogContent.innerHTML = "";
  self._processingDialogContent.style.textAlign = "center";
  self._processingDialogContent.appendChild(messageImg);

  $(self._processingDialog).dialog("option", "title", messageText);

  $(self._processingDialog).dialog("open");
};

GuiConnector.prototype.hideProcessing = function () {
  var self = returnThisOrSingleton(this);
  $(self._processingDialog).dialog("close");
};

GuiConnector.prototype.alert = function (error) {
  var self = returnThisOrSingleton(this);
  logger.error(error);
  if (self._errorDialog === undefined) {
    self._errorDialog = document.createElement("div");
    self._errorDialogContent = document.createElement("div");
    self._errorDialog.appendChild(self._errorDialogContent);
    document.body.appendChild(self._errorDialog);
    $(self._errorDialog).dialog({
      classes: {
        "ui-dialog": "ui-state-error"
      },
      modal: true,
      title: "ERROR"
    }).siblings('.ui-dialog-titlebar').css("background", "red");
  }
  var message = error;
  if (message instanceof SecurityError) {
    message = error.message + "<p>Please <a href=\"login.xhtml?from=" + encodeURI(window.location.href) + "\">login</a> to access this resource</p>";
  } else if (message instanceof ValidationError) {
    message = error.message;
  } else if (message instanceof GuiMessageError) {
    message = error.message;
  } else if (message instanceof NetworkError && (error.message === "XHR error" || error.message === "Failed to fetch")) {
    message = "File to big.<br>Please reduce file size or contact administrators.";
  } else if (message instanceof ValidationError) {
    message = error.message;
  } else if (message instanceof GuiMessageError) {
    message = error.message;
  } else if (message instanceof Error) {
    message = "Unexpected error occurred:<p>" + error.message + "</p>";
  }
  self._errorDialogContent.innerHTML = message;
  $(self._errorDialog).dialog("open");

};

GuiConnector.prototype.info = function (message) {
  var self = returnThisOrSingleton(this);
  if (self._infoDialog === undefined) {
    self._infoDialog = document.createElement("div");
    self._infoDialogContent = document.createElement("div");
    self._infoDialog.appendChild(self._infoDialogContent);
    document.body.appendChild(self._infoDialog);
    $(self._infoDialog).dialog({
      classes: {
        "ui-dialog": "ui-state-info"
      },
      modal: true,
      title: "INFO"
    });
  }
  self._infoDialogContent.innerHTML = message;
  $(self._infoDialog).dialog("open");

};

GuiConnector.prototype.destroy = function () {
  var self = returnThisOrSingleton(this);

  if (self._infoDialog !== undefined) {
    $(self._infoDialog).dialog("destroy").remove();
  }
  if (self._processingDialog !== undefined) {
    $(self._processingDialog).dialog("destroy").remove();
  }

  if (self._errorDialog !== undefined) {
    $(self._errorDialog).dialog("destroy").remove();
  }
};

GuiConnector.singleton = new GuiConnector();


module.exports = GuiConnector.singleton;
