"use strict";

var Promise = require("bluebird");

var logger = require('../logger');
var Functions = require('../Functions');

var AbstractCustomMap = require('./AbstractCustomMap');
var AbstractMarker = require('./marker/AbstractMarker');
var AbstractSurfaceElement = require('./surface/AbstractSurfaceElement');
var AliasMarker = require('./marker/AliasMarker');
var AliasSurface = require('./surface/AliasSurface');
var Alias = require('./data/Alias');
var CommentDialog = require('../gui/CommentDialog');
var ControlType = require('./ControlType');
var CustomMapOptions = require('./CustomMapOptions');
var GuiConnector = require('../GuiConnector');
var IdentifiedElement = require('./data/IdentifiedElement');
var LayoutData = require('./data/LayoutData');
var PointData = require('./data/PointData');
var PointMarker = require('./marker/PointMarker');
var Reaction = require('./data/Reaction');
var ReactionMarker = require('./marker/ReactionMarker');
var ReactionSurface = require('./surface/ReactionSurface');
var ReferenceGenome = require('./data/ReferenceGenome');
var SecurityError = require('../SecurityError');
var Submap = require('./Submap');
var TouchMap = require('./TouchMap');

/**
 * Default constructor.
 *
 * @param options
 *          CustomMapOptions object representing all parameters needed for map
 *          creation
 */
function CustomMap(options) {
  this._controls = [];

  if (!(options instanceof CustomMapOptions)) {
    options = new CustomMapOptions(options);
  }
  AbstractCustomMap.call(this, options.getProject().getModel(), options);
  this.setProject(options.getProject());

  this.registerListenerType("onBioEntityClick");

  this._selectedOverlays = [];

  this.customizeGoogleMapView(options.getElement());

  this.createMapChangedCallbacks();

  this.overlayCollections = [];

  // which submap is active (where user made interaction for the last time)
  this._activeSubmapId = null;

  this.initialized = true;

  // list of reference genomes
  this._referenceGenome = [];

  if (options.isCustomTouchInterface()) {
    this._touchInterface = new TouchMap(this);
  }

  this.createSubmaps();

  this._dialogs = [];
}

CustomMap.prototype = Object.create(AbstractCustomMap.prototype);
CustomMap.prototype.constructor = CustomMap;

/**
 * This code must be run after the object is created. It requires to download
 * some data via promises.
 *
 * @return promise with empty result
 */
CustomMap.prototype.init = function () {
  var self = this;
  return ServerConnector.getLogoImg().then(function (url) {
    var logo2 = self.getControl(ControlType.LOGO_2_IMG);
    if (!/^(f|ht)tps?:\/\//i.test(url)) {
      url = GuiConnector.getImgPrefix() + url;
    }
    logo2.src = url;
  });
};

/**
 * Create submaps.
 */
CustomMap.prototype.createSubmaps = function () {
  this.submaps = [];
  for (var i = 0; i < this.getModel().getSubmodels().length; i++) {
    this.submaps.push(new Submap(this, this.getModel().getSubmodels()[i]));
  }
};

/**
 * Creates logo and put it on the google map.
 */
CustomMap.prototype.createLogo = function () {
  var logoControlDiv2 = document.createElement('DIV');
  var logo2 = document.createElement('IMG');
  logo2.style.cursor = 'pointer';
  logo2.style.width = "80px";
  logo2.id = ControlType.LOGO_2_IMG;
  this.addControl(logo2);

  logoControlDiv2.appendChild(logo2);
  google.maps.event.addDomListener(logo2, 'click', function () {
    return ServerConnector.getLogoLink().then(function (url) {
      var win = window.open(url, '_blank');
      win.focus();
    });
  });
  logoControlDiv2.index = 0; // used for ordering
  this.getGoogleMap().controls[google.maps.ControlPosition.LEFT_BOTTOM].push(logoControlDiv2);

  var logoControlDiv = document.createElement('DIV');
  logoControlDiv.style.padding = '5px';

  var logo = document.createElement('IMG');
  logo.src = GuiConnector.getImgPrefix() + GuiConnector.getLcsbLogoImg(this.bigButtons);
  logo.style.cursor = 'pointer';
  logo.id = ControlType.LOGO_IMG;
  this.addControl(logo);

  logoControlDiv.appendChild(logo);
  google.maps.event.addDomListener(logo, 'click', function () {
    var win = window.open('http://wwwen.uni.lu/lcsb/', '_blank');
    win.focus();
  });

  logoControlDiv.index = 1; // used for ordering
  this.getGoogleMap().controls[google.maps.ControlPosition.RIGHT_BOTTOM].push(logoControlDiv);
};

/**
 * Clear all AbstractDbOverlay.
 */
CustomMap.prototype.clearDbOverlays = function () {
  var promises = [];
  for (var overlayName in this.overlayCollections) {
    if (this.overlayCollections.hasOwnProperty(overlayName)) {
      var collection = this.overlayCollections[overlayName];
      promises.push(collection.clear());
    }
  }
  return Promise.all(promises);
};

/**
 * Open data overlay given in the parameter.
 *
 * @param param
 *          identifier or LayoutData identifying data overlay to open
 */
CustomMap.prototype.openDataOverlay = function (param) {
  logger.debug("Opening data overlay: " + param);

  var self = this;
  var identifier = param;
  if (param instanceof LayoutData) {
    identifier = param.getId();
  }
  identifier = parseInt(identifier);

  if (isNaN(identifier)) {
    throw new Error("invalid id: " + param);
  }

  var index = null;
  var overlayToOpen = null;
  for (var j = 0; j < this.getLayouts().length; j++) {
    var overlay = this.getLayouts()[j];
    if (overlay.getId() === identifier) {
      index = j;
      overlayToOpen = overlay;
    }
  }

  if (index === null) {
    throw new SecurityError("You have no privileges for selected overlay");
  } else {
    return overlayToOpen.init().then(function () {
      if (overlayToOpen.getInputDataAvailable()) {
        if (self._selectedOverlays[identifier] === true) {
          logger.warn("Overlay " + identifier + " already selected");
          return Promise.resolve();
        } else {
          self._selectedOverlays[identifier] = true;
          return self.getVisibleDataOverlays().then(function (visibleDataOverlays) {
            var ids = [];
            for (var i = 0; i < visibleDataOverlays.length; i++) {
              ids.push(visibleDataOverlays[i].getId());
            }
            ServerConnector.getSessionData(self.getProject()).setVisibleOverlays(ids);
            return self.refreshSelectedLayouts();
          });

        }
      } else {
        self.getGoogleMap().setMapTypeId(identifier.toString());

        var submaps = self.getSubmaps();
        for (var i = 0; i < submaps.length; i++) {
          var submap = submaps[i];
          submap.openDataOverlay(submap.getModel().getLayouts()[index].getId());
        }
        return Promise.resolve();
      }
    });
  }
};

/**
 * Returns data overlay that is used as a background.
 */
CustomMap.prototype.getBackgroundDataOverlay = function () {
  var identifier = parseInt(this.getGoogleMap().getMapTypeId());
  for (var i = 0; i < this.getLayouts().length; i++) {
    var layout = this.getLayouts()[i];
    if (layout.getId() === identifier) {
      return layout;
    }
  }
  return null;
};

/**
 * Register AbstractDbOverlay in the map.
 *
 * @param dbOverlay
 *          database overlay to be connected to the map
 */
CustomMap.prototype.registerDbOverlay = function (dbOverlay) {
  var self = this;

  this.overlayCollections[dbOverlay.getName()] = dbOverlay;

  dbOverlay.markers = {
    ALIAS: [],
    REACTION: [],
    POINT: []
  };
  dbOverlay.mapOverlays = {
    ALIAS: [],
    REACTION: [],
    POINT: []
  };

  if (dbOverlay.getName() !== undefined) {
    if (dbOverlay.getName() === "search" || dbOverlay.getName().startsWith("plugin")) {
      dbOverlay.addListener("onSearch", function (e) {
        return self.renderOverlayCollection({
          overlayCollection: dbOverlay,
          fitBounds: e.arg.fitBounds
        });
      });
    }
    if (dbOverlay.getName() === "drug" || dbOverlay.getName() === "chemical" || dbOverlay.getName() === "mirna") {
      dbOverlay.addListener("onSearch", function (e) {
        return self.renderOverlayCollection({
          overlayCollection: dbOverlay,
          fitBounds: e.arg.fitBounds
        });
      });

      dbOverlay.addListener("onTargetVisibilityChange", function () {
        return self.renderOverlayCollection({
          overlayCollection: dbOverlay,
          fitBounds: false
        });
      });
    }
    if (dbOverlay.getName() === "comment") {
      var listener = function () {
        return self.renderOverlayCollection({
          overlayCollection: dbOverlay,
          fitBounds: false
        });
      };
      dbOverlay.addListener("onRefresh", listener);
      dbOverlay.addListener("onClear", listener);
    }
  }

};

/**
 * Refresh comment list.
 *
 * @return promise that is resolved when comment list is refreshed
 */
CustomMap.prototype.refreshComments = function () {
  var self = this;

  var commentDbOverlay = self.getOverlayByName("comment");
  if (commentDbOverlay !== undefined) {
    if (ServerConnector.getSessionData(self.getProject()).getShowComments()) {
      return commentDbOverlay.refresh();
    } else {
      return commentDbOverlay.clear();
    }
  }
  throw new Error("comment DbOverlay not found");
};

/**
 * Turns on/off drawing module on google maps.
 */
CustomMap.prototype.turnOnOffDrawing = function () {
  var model = this.getSubmapById(this.getActiveSubmapId());
  if (model !== undefined) {
    model._turnOnOffDrawing();
  } else {
    throw new Error("Cannot find submodel with id: " + this.getActiveSubmapId());
  }
};

CustomMap.prototype.refreshMarkers = function (force) {
  logger.debug("Refresh Markers: ");
  for (var overlayName in this.overlayCollections) {
    if (this.overlayCollections.hasOwnProperty(overlayName)) {
      var collection = this.overlayCollections[overlayName];
      this.refreshOverlayMarkers(collection, force);
    }
  }
};

CustomMap.prototype.refreshOverlayMarkers = function (overlay, force) {
  logger.debug("Refresh overlay: " + overlay.name);

  if (!this.isMarkerOptimization() || force) {
    for (var markerType in overlay.markers) {
      if (overlay.markers.hasOwnProperty(markerType)) {
        var markers = overlay.markers[markerType];
        for (var key in markers) {
          if (markers.hasOwnProperty(key)) {
            var marker = markers[key];
            marker.hide();
            marker.show();
          }
        }
      }
    }
  }
};

/**
 * Returns HTML div that is attached to jQuery dialog that should be used for
 * opening submap.
 *
 * @param id
 *          identifier of the submap
 */
CustomMap.prototype.getSubmapDialogDiv = function (id) {
  var dialogDiv = this._dialogs[id];

  if (dialogDiv === undefined) {
    dialogDiv = document.createElement("div");
    dialogDiv.setAttribute("name", "dialog-" + id);

    this._dialogs[id] = dialogDiv;

    $(dialogDiv).dialog({
      autoOpen: false
    });
  }
  return dialogDiv;
};

/**
 * Opens a dialog with a submodel.
 */
CustomMap.prototype.openSubmap = function (id) {
  var self = this;
  var submap = self.getSubmapById(id);
  if (submap === null) {
    throw new Error("Unknown submap for id: " + id);
  } else if (submap !== this) {
    var wasInitialized = submap.initialized;

    var dialogDiv = self.getSubmapDialogDiv(id);

    submap.open(dialogDiv);

    // we have to perform it on top map, because on submaps id is different
    return this.openDataOverlay(self.getGoogleMap().getMapTypeId()).then(function () {
      // now we have to visualize layouts
      var layouts = [];

      // get list of layouts
      for (var key in self._selectedOverlays) {
        if (self._selectedOverlays.hasOwnProperty(key) && self._selectedOverlays[key] === true) {
          layouts.push(key);
        }
      }

      // show layouts that should be visualized (resize or show them)
      for (var i = 0; i < layouts.length; i++) {
        var layoutId = layouts[i];
        submap._showSelectedLayout(layoutId, i, layouts.length);
      }
      if (!wasInitialized) {
        return self.refreshMarkers(true);
      }
    });
  }

};

CustomMap.prototype.customizeGoogleMapView = function (div) {
  var mapOptions = this.createMapOptions();
  this.setGoogleMap(new google.maps.Map(div, mapOptions));
  this.setupLayouts();

  this.registerMapClickEvents();

  this.createLogo();
  // this.createMapVersion();
  google.maps.event.trigger(this.getGoogleMap(), 'resize');
  google.maps.event.trigger(this.getGoogleMap(), 'maptypeid_changed');
  google.maps.event.trigger(this.getGoogleMap(), 'projection_changed');

  // center map and zoom in to fit into browser window if there is no
  // information about coordinates in the session
  if (ServerConnector.getSessionData(this.getProject()).getCenter(this.getModel()) === undefined) {
    var bounds = new google.maps.LatLngBounds();
    bounds.extend(this.getTopLeftLatLng());
    bounds.extend(this.getBottomRightLatLng());

    this.getGoogleMap().fitBounds(bounds);
  }
};

/**
 * Creates listeners for google.maps.Map object that will actualize the data in
 * user session.
 */
CustomMap.prototype.createMapChangedCallbacks = function () {
  this._createMapChangedCallbacks();
  var self = this;
  var sessionData = ServerConnector.getSessionData(self.getProject());

  // if we have zoom level data stored in session then restore it
  var level = sessionData.getZoomLevel(self.getModel());
  if (parseInt(level) > 0) {
    level = parseInt(level);
    this.getGoogleMap().setZoom(level);
  } else {
    sessionData.setZoomLevel(self.getModel(), self.getGoogleMap().getZoom());
  }

  // listener for changing type of layout
  google.maps.event.addListener(self.getGoogleMap(), 'maptypeid_changed', function () {
    sessionData.setSelectedBackgroundOverlay(self.getGoogleMap().getMapTypeId());
  });

  // if we have type of layout stored in the session then restore it
  var mapType = sessionData.getSelectedBackgroundOverlay();
  if (mapType !== undefined) {
    try {
      self.openDataOverlay(mapType);
    } catch (e) {
      if (e instanceof SecurityError) {
        logger.debug(e.message);
        sessionData.setSelectedBackgroundOverlay(self.getLayouts()[0].getId());
      } else {
        throw e;
      }
    }
  }
};

/**
 * Returns submap (or this map) by id.
 *
 * @param identifier
 *          identifier of the submap
 * @returns AbstractCustomMap (submap or this map) with given identifier of the model
 */
CustomMap.prototype.getSubmapById = function (identifier) {
  identifier = parseInt(identifier);
  if (this.getId() === identifier) {
    return this;
  }
  for (var i = 0; i < this.submaps.length; i++) {
    if (this.submaps[i].getId() === identifier) {
      return this.submaps[i];
    }
  }
  logger.warn("Cannot find submodel with id: " + identifier);
  return null;
};

CustomMap.prototype.removeSelection = function () {
  var model = this.getSubmapById(this.getActiveSubmapId());
  if (model !== null) {
    model._removeSelection();
  } else {
    throw new Error("Cannot find submap with id: " + this.getActiveSubmapId());
  }
};

/**
 * Removes layout from visualization.
 *
 * @param identifier
 *          identifier of layout to remove
 *
 */
CustomMap.prototype.removeSelectedLayout = function (identifier) {
  var self = this;
  logger.debug("Removing layout: " + identifier);

  if (this._selectedOverlays[identifier] !== true) {
    logger.warn("Overlay " + identifier + " is not selected");
    return Promise.resolve();
  } else {
    self._selectedOverlays[identifier] = false;
    return self.refreshSelectedLayouts().then(function () {
      return self.getVisibleDataOverlays();
    }).then(function (visibleDataOverlays) {
      var ids = [];
      for (var i = 0; i < visibleDataOverlays.length; i++) {
        ids.push(visibleDataOverlays[i].getId());
      }
      ServerConnector.getSessionData(self.getProject()).setVisibleOverlays(ids);
      return self.refreshSelectedLayouts();
    });
  }
};

/**
 * Refresh visualization of selected layouts.
 */
CustomMap.prototype.refreshSelectedLayouts = function () {
  logger.debug("Refreshing layouts");
  var self = this;
  return self.getVisibleDataOverlays().then(function (visibleDataOverlays) {
    // show layouts that should be visualized (resize or show them)
    var promises = [];
    for (var i = 0; i < visibleDataOverlays.length; i++) {
      var layoutId = visibleDataOverlays[i].getId();
      if (self.layoutContainsOverlays(layoutId)) {
        // resize element on the map
        promises.push(self.resizeSelectedLayout(layoutId, i, visibleDataOverlays.length));
      } else {
        promises.push(self.showSelectedLayout(layoutId, i, visibleDataOverlays.length));
      }
    }
    return Promise.all(promises).then(function () {
      // remove layouts that were
      for (var key in self.selectedLayoutOverlays) {
        if (!self._selectedOverlays.hasOwnProperty(key) || self._selectedOverlays[key] === false) {
          if (self.layoutContainsOverlays(key)) {
            self.hideSelectedLayout(key);
          }
        }
      }
      return self.refreshInfoWindows();
    });
  });
};

/**
 * Hides layout from the map and all submaps
 *
 * @param layoutId
 *          identifier of a layout to hide
 */
CustomMap.prototype.hideSelectedLayout = function (layoutId) {
  this._hideSelectedLayout(layoutId);
  for (var i = 0; i < this.submaps.length; i++) {
    this.submaps[i]._hideSelectedLayout(layoutId);
  }
};

/**
 * Resize(refresh) layout on the map and all submaps. Resizing should be called
 * when number of layouts to visualize change.
 *
 * @param layoutId
 *          identifier of layout to refresh
 * @param index
 *          position of the layout in list of layouts that we visualize
 * @param length
 *          number of layouts that we currently visualize
 */
CustomMap.prototype.resizeSelectedLayout = function (layoutId, index, length) {
  logger.debug("Resize layout: " + layoutId);
  var promises = [];
  promises.push(this._resizeSelectedLayout(layoutId, index, length));
  for (var i = 0; i < this.submaps.length; i++) {
    promises.push(this.submaps[i]._resizeSelectedLayout(layoutId, index, length));
  }
  return Promise.all(promises);
};

/**
 * Show layout on the map and all submaps.
 *
 * @param layoutId
 *          identifier of layout to show
 * @param index
 *          position of the layout in list of layouts that we visualize
 * @param length
 *          number of layouts that we currently visualize
 */
CustomMap.prototype.showSelectedLayout = function (layoutId, index, length) {
  logger.debug("Resize layout: " + layoutId);
  var promises = [];
  promises.push(this._showSelectedLayout(layoutId, index, length));

  for (var i = 0; i < this.submaps.length; i++) {
    promises.push(this.submaps[i]._showSelectedLayout(layoutId, index, length));
  }
  return Promise.all(promises);
};

/**
 * This method checks if the layout contains any overlays (like AliasSurface or
 * ReactionSurface) that is currently visible on the map.
 *
 * @param layoutId
 *          identifier of the layout
 * @returns {Boolean}: <code>true</code> if the layout contains overlays to
 *          visualize, <code>false</code> otherwise
 */
CustomMap.prototype.layoutContainsOverlays = function (layoutId) {

  // first, check top map
  if (this.selectedLayoutOverlays.hasOwnProperty(layoutId) && this.selectedLayoutOverlays[layoutId].length > 0) {
    return true;
  }

  // now check all submaps
  for (var i = 0; i < this.submaps.length; i++) {
    if (this.submaps[i].initialized) {
      if (this.submaps[i].selectedLayoutOverlays.hasOwnProperty(layoutId)
        && this.submaps[i].selectedLayoutOverlays[layoutId].length > 0) {
        return true;
      }
    }
  }
  return false;
};

/**
 * Refresh content of all {@link AliasInfoWindow} in this map and all submaps.
 */
CustomMap.prototype.refreshInfoWindows = function () {
  var promises = [];
  promises.push(this._refreshInfoWindows());
  // now check all submaps
  for (var i = 0; i < this.submaps.length; i++) {
    promises.push(this.submaps[i]._refreshInfoWindows());
  }
  return Promise.all(promises);
};

/**
 * Opens {@link AliasInfoWindow} for an {@link Alias} in a given model/submodel.
 *
 * @param aliasId
 *          identifier of {@link Alias}
 * @param modelId
 *          identifier of {@link AbstractCustomMap}
 */
CustomMap.prototype.openInfoWindowForAlias = function (alias, marker) {
  logger.debug("Opening info window for alias: " + alias.getId() + ", model: " + alias.getModelId());
  var submap = this.getSubmapById(alias.getModelId());
  return submap._openInfoWindowForAlias(alias, marker);
};

/**
 * Renders markers, lines, etc. for elements highlighted in OverlayCollection.
 *
 * @param overlayCollection
 *          {@link OverlayCollection} to be processed
 * @param fitBounds
 *          <code>true</code> if the borders should fit bounds after creating
 *          all elements
 */
CustomMap.prototype.renderOverlayCollection = function (params) {
  var self = this;
  var fitBounds = params.fitBounds;
  var overlayCollection = params.overlayCollection;

  var elements;
  var markers = [];

  return overlayCollection.getIdentifiedElements().then(function (identifiedElements) {
    elements = identifiedElements;

    return Promise.each(elements, function (element) {
      var icon = element.getIcon();
      if (icon !== null && icon !== undefined) {
        return self.createMarkerForDbOverlay(element, overlayCollection).then(function (marker) {

          markers.push(marker);
        });
      } else {
        return self.createSurfaceForDbOverlay(element, overlayCollection).then(function (mapOverlay) {
          markers.push(mapOverlay);
        });
      }
    });
  }).then(function () {
    self.removeUnmodifiedMarkersAndSurfaces(markers, overlayCollection);

    return Promise.each(elements, function (element) {
      var infoWindow = self.getInfoWindowForIdentifiedElement(element);
      if (infoWindow !== null && infoWindow !== undefined) {
        return self.retrieveOverlayDetailDataForElement(element, infoWindow.getOverlayFullViewArray()).then(function () {
          return self.updateInfoWindowForIdentifiedElement(element);
        });
      }
    });
  }).then(function () {
    if (elements.length > 0 && fitBounds) {
      self.fitBounds(markers);
      for (var j = 0; j < self.submaps.length; j++) {
        self.submaps[j].fitBounds(markers);
      }
    }
  });
};

CustomMap.prototype.removeUnmodifiedMarkersAndSurfaces = function (modifiedMarkersAndSurfaces, dbOverlay) {
  var modifiedMarkers = {
    "ALIAS": [],
    "REACTION": [],
    "POINT": [],
  };
  var modifiedSurfaces = {
    "ALIAS": [],
    "REACTION": [],
    "POINT": [],
  };

  for (var i = 0; i < modifiedMarkersAndSurfaces.length; i++) {
    var object = modifiedMarkersAndSurfaces[i];
    var identifiedElement = object.getIdentifiedElement();
    if (object instanceof AbstractSurfaceElement) {
      modifiedSurfaces[identifiedElement.getType()][identifiedElement.getId()] = true;
    } else if (object instanceof AbstractMarker) {
      modifiedMarkers[identifiedElement.getType()][identifiedElement.getId()] = true;
    } else {
      throw new Error("Unknown class type: " + object.prototype.name);
    }
  }

  var markerType, key;
  for (markerType in dbOverlay.markers) {
    if (dbOverlay.markers.hasOwnProperty(markerType)) {
      var markers = dbOverlay.markers[markerType];
      for (key in markers) {
        if (markers.hasOwnProperty(key) && !modifiedMarkers[markerType][key]) {
          markers[key].setMap(null);
          delete markers[key];
        }
      }
    }
  }

  for (markerType in dbOverlay.mapOverlays) {
    if (dbOverlay.mapOverlays.hasOwnProperty(markerType)) {
      var mapOverlays = dbOverlay.mapOverlays[markerType];
      for (key in mapOverlays) {
        if (mapOverlays.hasOwnProperty(key) && !modifiedSurfaces[markerType][key]) {
          mapOverlays[key].setMap(null);
          delete mapOverlays[key];
        }
      }
    }
  }
};

/**
 * Opens {@link AbstractInfoWindow} for a marker.
 *
 * @param marker
 *          {@link AbstractMarker} for which info window should be opened
 */
CustomMap.prototype.openInfoWindowForIdentifiedElement = function (element, marker) {
  var self = this;
  var submap = self.getSubmapById(element.getModelId());
  logger.debug(element + ": Opening info window");

  // open AliasInfoWindow in a right model
  return submap._openInfoWindowForIdentifiedElement(element, marker).then(function () {
    var infoWindow = submap.returnInfoWindowForIdentifiedElement(element);
    return self.retrieveOverlayDetailDataForElement(element, infoWindow.getOverlayFullViewArray());
  });

};

/**
 * Sends requests to download detailed data about element in all
 * {@link OverlayCollection}.
 *
 * @param element
 *          element for which we want to have detailed information
 */
CustomMap.prototype.retrieveOverlayDetailDataForElement = function (element, general) {
  var promises = [];
  if (general === undefined) {
    logger.warn("general param is undefined!");
    general = [];
  }
  for (var overlayName in this.overlayCollections) {
    if (this.overlayCollections.hasOwnProperty(overlayName)) {
      var overlay = this.overlayCollections[overlayName];

      var generalRequest = general[overlayName];
      if (generalRequest === undefined) {
        logger.warn("No information about general overlay request for overlay: ", overlayName);
        generalRequest = false;
      }
      generalRequest = generalRequest || (!overlay.allowSearchById() && overlay.allowGeneralSearch());

      promises.push(overlay.getDetailDataByIdentifiedElement(element, generalRequest));
    }
  }
  return Promise.all(promises);
};

/**
 * Updates info window identified by element given as a parameter.
 *
 * @param identifiedElement
 *          element for which info window should be updated
 */
CustomMap.prototype.updateInfoWindowForIdentifiedElement = function (identifiedElement) {
  var infoWindow = this.getInfoWindowForIdentifiedElement(identifiedElement);
  if (infoWindow === null || infoWindow === undefined) {
    return Promise.resolve();
  } else {
    return infoWindow.update();
  }
};

/**
 * Returns data from all {@link OverlayCollection} for a given alias.
 *
 * @param alias
 *          {@link Alias} for which overlay data will be returned
 * @param general
 *          if true then all elements will be returned, if false then only ones
 *          available right now in the overlay
 * @returns data from all {@link OverlayCollection} for a given alias
 */
CustomMap.prototype.getOverlayDataForAlias = function (alias, general) {
  var identifiedElement = new IdentifiedElement(alias);
  return this.getOverlayDataForIdentifiedElement(identifiedElement, general);
};

/**
 * Returns data from all {@link OverlayCollection} for a given reaction.
 *
 * @param reaction
 *          {@link Reaction} for which overlay data will be returned
 * @param general
 *          if true then all elements will be returned, if false then only ones
 *          available right now in the overlay
 * @returns data from all {@link OverlayCollection} for a given alias
 */
CustomMap.prototype.getOverlayDataForReaction = function (reaction, general) {
  var identifiedElement = new IdentifiedElement(reaction);
  return this.getOverlayDataForIdentifiedElement(identifiedElement, general);
};

/**
 * Returns data from all {@link OverlayCollection} for a given {@link PointData}
 *
 * @param point
 *          {@link PointData} for which overlay data will be returned
 * @returns data from all {@link OverlayCollection} for a given
 *          {@link PointData}
 */
CustomMap.prototype.getOverlayDataForPoint = function (point, general) {
  var identifiedElement = new IdentifiedElement(point);
  return this.getOverlayDataForIdentifiedElement(identifiedElement, general);
};

/**
 * Returns data from all {@link OverlayCollection} for element identified by the
 * parameter
 *
 * @param identifiedElement
 *          {@link IdentifiedElement} for which overlay data will be returned
 * @returns data from all {@link OverlayCollection} for a given
 *          {@link IdentifiedElement}
 */
CustomMap.prototype.getOverlayDataForIdentifiedElement = function (identifiedElement, general) {
  if (general === undefined) {
    logger.warn("general parameter must be defined");
    general = [];
  }
  var promises = [];
  var overlays = [];
  for (var overlayName in this.overlayCollections) {
    if (this.overlayCollections.hasOwnProperty(overlayName)) {
      var overlay = this.overlayCollections[overlayName];
      if (overlay.allowGeneralSearch() || overlay.allowSearchById()) {
        var generalFlag = general[overlay.getName()];
        if (generalFlag === undefined) {
          logger.warn("General flag for overlay: " + overlay.getName() + " is not defined, assuming false");
          generalFlag = false;
        }
        overlays.push(overlay);
        promises.push(overlay.getDetailDataByIdentifiedElement(identifiedElement, !overlay.allowSearchById()
          || generalFlag));
      }
    }
  }
  return Promise.all(promises).then(function (values) {
    var result = [];
    for (var i = 0; i < values.length; i++) {
      result.push({
        overlay: overlays[i],
        data: values[i]
      });
    }
    return result;
  });
};

CustomMap.prototype.getOverlayByName = function (name) {
  return this.overlayCollections[name];
};

CustomMap.prototype.getDbOverlays = function () {
  var result = [];
  for (var overlayName in this.overlayCollections) {
    if (this.overlayCollections.hasOwnProperty(overlayName)) {
      result.push(this.overlayCollections[overlayName]);
    }
  }
  return result;
};

/**
 * Returns {@link AbstractInfoWindow} for element identified by the parameter.
 *
 * @param identifiedElement
 *          {@link IdentifiedElement} that determines for which element we want
 *          {@link AbstractInfoWindow}
 * @returns {@link AbstractInfoWindow} for element identified by the parameter
 */
CustomMap.prototype.getInfoWindowForIdentifiedElement = function (identifiedElement) {
  var model = this.getSubmapById(identifiedElement.modelId);
  var infoWindow = null;
  if (identifiedElement.type === "ALIAS") {
    infoWindow = model.getAliasInfoWindowById(identifiedElement.getId());
  } else if (identifiedElement.type === "POINT") {
    infoWindow = model.getPointInfoWindowById(identifiedElement.getId());
  } else if (identifiedElement.type === "REACTION") {
    infoWindow = model.getReactionInfoWindowById(identifiedElement.getId());
  } else {
    throw new Error("Unknown type of IdentifiedElement: ", identifiedElement);
  }
  return infoWindow;
};

CustomMap.prototype.getActiveSubmapId = function () {
  return this._activeSubmapId;
};

CustomMap.prototype.setActiveSubmapId = function (submapId) {
  this._activeSubmapId = submapId;
};

CustomMap.prototype.setActiveSubmapClickCoordinates = function (coordinates) {
  if (!(coordinates instanceof google.maps.Point)) {
    throw new Error("Coordinates must be provided as google.maps.Point object, but found: " + coordinates);
  }
  this._activeSubmapCoordinates = coordinates;
};
CustomMap.prototype.getActiveSubmapClickCoordinates = function () {
  return this._activeSubmapCoordinates;
};

CustomMap.prototype.getReferenceGenome = function (type, version) {
  var self = this;
  if (self.getProject().getOrganism() === undefined) {
    return Promise.resolve(new ReferenceGenome(null));
  }

  if (self._referenceGenome[type] === undefined) {
    self._referenceGenome[type] = [];
  }
  if (self._referenceGenome[type][version] === undefined) {
    return ServerConnector.getReferenceGenome({
      type: type,
      version: version,
      organism: self.getProject().getOrganism().getResource(),
    }).then(function (genome) {
      self._referenceGenome[type][version] = genome;
      return genome;
    });
  } else {
    return Promise.resolve(self._referenceGenome[type][version]);
  }
};

CustomMap.prototype.getTopMap = function () {
  return this;
};

CustomMap.prototype.setProject = function (project) {
  this._project = project;
};
CustomMap.prototype.getProject = function () {
  return this._project;
};

CustomMap.prototype.getTopOverviewImage = function () {
  return this.getProject().getTopOverviewImage();
};

CustomMap.prototype.addControl = function (element) {
  if (this._controls[element.id] !== undefined) {
    throw new Error("Element with id " + element.id + " already added to controls");
  }
  if (ControlType[element.id] === undefined) {
    throw new Error("Unknown control type: " + element.id);
  }
  this._controls[element.id] = element;
};

CustomMap.prototype.getControl = function (type) {
  if (ControlType[type] === undefined) {
    throw new Error("Unknown control type: " + type);
  }
  return this._controls[type];
};

CustomMap.prototype.setCommentDialog = function (commentDialog) {
  this._commentDialog = commentDialog;
};

CustomMap.prototype.getCommentDialog = function () {
  return this._commentDialog;
};

CustomMap.prototype.setContextMenu = function (contextMenu) {
  this._contextMenu = contextMenu;
};

CustomMap.prototype.getContextMenu = function () {
  return this._contextMenu;
};

CustomMap.prototype.setSelectionContextMenu = function (selectionContextMenu) {
  this._selectionContextMenu = selectionContextMenu;
};

CustomMap.prototype.getSelectionContextMenu = function () {
  return this._selectionContextMenu;
};

CustomMap.prototype.openCommentDialog = function () {
  var self = this;
  return ServerConnector.getClosestElementsByCoordinates({
    modelId: this.getActiveSubmapId(),
    coordinates: this.getActiveSubmapClickCoordinates()
  }).then(function (elements) {
    var commentDialog = self.getCommentDialog();
    if (commentDialog === undefined) {
      var div = Functions.createElement({
        type: "div"
      });
      self.getElement().appendChild(div);
      commentDialog = new CommentDialog({
        element: div,
        customMap: self
      });
      self.setCommentDialog(commentDialog);
    }
    self.setCommentDialog(commentDialog);

    return commentDialog.open(elements);
  });
};

CustomMap.prototype.fetchIdentifiedElements = function (elements, complete) {
  var modelIds = [];
  var modelElements = [];
  var i;
  for (i = 0; i < elements.length; i++) {
    var element = elements[i];
    if (modelElements[element.getModelId()] === undefined) {
      modelIds.push(element.getModelId());
      modelElements[element.getModelId()] = [];
    }
    modelElements[element.getModelId()].push(element);
  }

  var promises = [];
  for (i = 0; i < modelIds.length; i++) {
    var modelId = modelIds[i];
    promises.push(this.getSubmapById(modelId).getModel().getByIdentifiedElements(modelElements[modelId], complete));
  }
  return Promise.all(promises);

};

CustomMap.prototype.refreshSearchAutocomplete = function () {
  var self = this;
  self._searchAutocomplete = [];
  return ServerConnector.getSuggestedQueryList().then(function (queries) {
    for (var i = 0; i < queries.length; i++) {
      var mainString = queries[i];
      for (var j = 0; j < mainString.length; j++) {
        var substring = mainString.substring(0, j + 1);
        if (self._searchAutocomplete[substring] !== undefined) {
          continue;
        }

        var list = [];
        for (var k = 0; k < 5; k++) {
          if (k + i >= queries.length) {
            break;
          } else if (queries[k + i].toLowerCase().startsWith(substring.toLowerCase())) {
            list.push(queries[k + i]);
          }
        }

        self._searchAutocomplete[substring] = list;
      }
    }
    return self._searchAutocomplete;
  });
};

CustomMap.prototype.getSearchAutocomplete = function (query) {
  if (this._searchAutocomplete === undefined) {
    this.refreshSearchAutocomplete();
    return [];
  }

  return this._searchAutocomplete[query];
};

CustomMap.prototype.setSelectedPolygon = function (polygonData) {
  this._selectedPolygon = polygonData;
};

CustomMap.prototype.getSelectedPolygon = function () {
  return this._selectedPolygon;
};

CustomMap.prototype.getDistance = function (params) {
  var self = this;
  var ie = params.element;
  var model = self.getSubmapById(ie.getModelId()).getModel();
  if (ie.getModelId() !== params.modelId) {
    throw new Error("Element and coordinates are on different maps: " + ie.getModelId() + ", " + params.modelId);
  }
  var x = params.coordinates.x;
  var y = params.coordinates.y;
  var p1 = new google.maps.Point(x, y);
  return model.getByIdentifiedElement(ie).then(function (element) {
    if (element instanceof Alias) {
      if (element.getX() <= x && element.getX() + element.getWidth() >= x) {
        if (element.getY() <= y && element.getY() + element.getHeight() >= y) {
          return 0;
        } else {
          return Math.min( //
            Math.abs(element.getY() - y), //
            Math.abs(element.getY() + element.getHeight() - y) //
          );
        }
      } else if (element.getY() <= y && element.getY() + element.getHeight() >= y) {
        return Math.min( //
          Math.abs(element.getX() - x), //
          Math.abs(element.getX() + element.getWidth() - x) //
        );
      } else {
        var elementX = element.getX();
        var elementY = element.getY();
        var elementWidth = element.getWidth();
        var elementHeight = element.getHeight();
        return Math.min( //
          Functions.distance(p1, new google.maps.Point(elementX, y)), //
          Functions.distance(p1, new google.maps.Point(elementX + elementWidth, elementY)), //
          Functions.distance(p1, new google.maps.Point(elementX, elementY + elementHeight)), //
          Functions.distance(p1, new google.maps.Point(elementX + elementWidth, elementY + elementHeight)) //
        );
      }
    } else if (element instanceof Reaction) {
      var distance = Number.POSITIVE_INFINITY;
      var lines = element.getLines();
      for (var i = 0; i < lines.length; i++) {
        distance = Math.min(distance, Functions.distance(p1, lines[i]));
      }
      return distance;
    } else {
      throw new Error("Unknown element type: " + (typeof element));
    }
  });
};

CustomMap.prototype.getSubmaps = function () {
  var submaps = this.submaps;
  if (submaps === undefined) {
    submaps = [];
  }
  return submaps;
};

CustomMap.prototype.destroy = function () {
  var self = this;
  var commentDialog = self.getCommentDialog();
  if (commentDialog !== undefined) {
    commentDialog.destroy();
  }
  var submaps = self.getSubmaps();
  for (var i = 0; i < submaps.length; i++) {
    submaps[i].destroy();
  }
};

CustomMap.prototype.getVisibleDataOverlays = function () {
  var dataOverlayPromises = [];

  // get list of layouts
  for (var key in this._selectedOverlays) {
    if (this._selectedOverlays.hasOwnProperty(key) && this._selectedOverlays[key] === true) {
      dataOverlayPromises.push(this.getModel().getLayoutDataById(key));
    }
  }

  return Promise.all(dataOverlayPromises);
};

CustomMap.prototype.createMarkerForDbOverlay = function (element, dbOverlay) {
  var self = this;

  var result = dbOverlay.markers[element.getType()][element.getId()];
  if (result !== undefined) {
    result.updateIdentifiedElement(element);
    return Promise.resolve(result);
  }

  var submap = self.getSubmapById(element.getModelId());
  var MarkerConstructor = null;
  if (element.getType() === "ALIAS") {
    MarkerConstructor = AliasMarker;
  } else if (element.getType() === "REACTION") {
    MarkerConstructor = ReactionMarker;
  } else if (element.getType() === "POINT") {
    MarkerConstructor = PointMarker;
  } else {
    throw new Error("Unknown type of the element in overlay: " + element.type);
  }
  result = new MarkerConstructor({
    element: element,
    map: submap,
    onClick: [function () {
      return self._openInfoWindowForIdentifiedElement(element, result.getGoogleMarker());
    }, function () {
      return self.callListeners("onBioEntityClick", element);
    }]
  });
  return result.init().then(function () {
    dbOverlay.markers[element.getType()][element.getId()] = result;
    return result;
  });
};

CustomMap.prototype.createSurfaceForDbOverlay = function (element, dbOverlay) {
  var self = this;

  var result = dbOverlay.mapOverlays[element.getType()][element.getId()];
  if (result !== undefined) {
    result.updateIdentifiedElement(element);
    return Promise.resolve(result);
  }

  var map = self.getSubmapById(element.getModelId());
  if (element.getType() === "ALIAS") {
    return AliasSurface.createFromIdentifiedElement({
      element: element,
      map: self,
      onClick: [function () {
        return self.openInfoWindowForIdentifiedElement(element, result.getGoogleMarker());
      }, function () {
        return self.callListeners("onBioEntityClick", element);
      }]
    }).then(function (surface) {
      result = surface;
      dbOverlay.mapOverlays[element.getType()][element.getId()] = result;
      return result;
    });
  } else if (element.getType() === "REACTION") {
    return map.getModel().getReactionById(element.getId()).then(function (reactionData) {
      var result = new ReactionSurface({
        reaction: reactionData,
        map: map,
        customized: true,
        color: element.getColor(),
        onClick: [function () {
          return self.openInfoWindowForIdentifiedElement(element, result.getGoogleMarker());
        }, function () {
          return self.callListeners("onBioEntityClick", element);
        }]
      });
      result.show();
      dbOverlay.mapOverlays[element.getType()][element.getId()] = result;
      return result;
    });
  } else if (element.getType() === "POINT") {
    throw new Error("Not implemented");
  } else {
    throw new Error("Unknown type of the element in overlay: " + element.type);
  }
};

/**
 * Opens {@link AbstractInfoWindow} for a marker.
 *
 * @param marker
 *          marker for which we are opening window
 */
CustomMap.prototype._openInfoWindowForIdentifiedElement = function (element, googleMarker) {
  var self = this;
  var submap = self.getSubmapById(element.getModelId());
  if (element.getType() === "ALIAS") {
    return submap.getModel().getByIdentifiedElement(element).then(function (alias) {
      return submap._openInfoWindowForAlias(alias, googleMarker);
    });
  } else if (element.getType() === "POINT") {
    return submap._openInfoWindowForPoint(new PointData(element), googleMarker);
  } else if (element.getType() === "REACTION") {
    return submap.getModel().getByIdentifiedElement(element).then(function (reaction) {
      return submap._openInfoWindowForReaction(reaction, googleMarker);
    });
  } else {
    throw new Error("Unknown element type: " + element.getType());
  }
};

module.exports = CustomMap;
